from functools import partial
from typing import Callable

import bpy

SUFFIX_TO_READER = {
    ".gltf": bpy.ops.import_scene.gltf,
    ".glb": bpy.ops.import_scene.gltf,
    # ".stl": bpy.ops.wm.stl_import,
    ".ply": bpy.ops.import_scene.ply,
    # ".obj": bpy.ops.wm.obj_import,
}


def export_scene_wrapper(exporter: Callable):
    return partial(exporter, use_selection=True)


glb_gltf_export = export_scene_wrapper(bpy.ops.export_scene.gltf)

SUFFIX_TO_WRITER = {
    ".gltf": glb_gltf_export,
    ".glb": glb_gltf_export,
    # ".stl": bpy.ops.wm.stl_export,
    ".ply": export_scene_wrapper(bpy.ops.export_scene.ply),
    # ".obj": bpy.ops.wm.obj_export,
}



class BlenderNotInstalledError(ModuleNotFoundError):
    pass


class Open3DNotInstalledError(ModuleNotFoundError):
    pass


class PyMeshNotInstalledError(ModuleNotFoundError):
    pass


def blender_module_not_installed_error(e):
    raise BlenderNotInstalledError(
        """
        Blender module, bpy, must be installed.
        
        Please run `pip install meshemy[blender]`
        """
    ) from e


def open3d_module_not_installed_error(e):
    raise Open3DNotInstalledError(
        """
        Open3D module must be installed.

        Please run `pip install meshemy[open3d]`
        """
    ) from e


def pymesh_module_not_installed_error(e):
    raise PyMeshNotInstalledError(
        """
        PyMesh module must be installed.

        Please run `pip install meshemy[pymesh]`
        """
    ) from e



from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar, Generic, ClassVar, Callable, Any

from pydantic import BaseModel, FilePath
from pydantic.generics import GenericModel
from pydantic_numpy import NDArray, NDArrayFp64

if TYPE_CHECKING:
    from meshemy.cookbook.blender import BlenderCookbook
    from meshemy.cookbook.open3d import Open3dCookbook


class BaseCookbook(BaseModel, ABC):
    """
    The .from_data() and .from_file() class method also needs to be implemented, I did not make it an abstractmethod
    because each Cookbook has its own function signature.
    """

    class Config:
        arbitrary_types_allowed = True
        keep_untouched = (cached_property,)

    @property
    @abstractmethod
    def vertices_numpy_array(self) -> NDArrayFp64 | None:
        ...

    @property
    @abstractmethod
    def edges_numpy_array(self) -> NDArray | None:
        ...

    @property
    @abstractmethod
    def faces_numpy_array(self) -> NDArray | None:
        ...

    @abstractmethod
    def save(self, save_path: Path | str) -> None:
        ...

    def to_o3d(self) -> "Open3dCookbook":
        from meshemy.cookbook.open3d import Open3dCookbook

        if isinstance(self, Open3dCookbook):
            return self

        return Open3dCookbook.from_data(self.vertices_numpy_array, self.faces_numpy_array)

    def to_blender(self, name: str) -> "BlenderCookbook":
        from meshemy.cookbook.blender import BlenderCookbook

        if isinstance(self, BlenderCookbook):
            return self

        return BlenderCookbook.from_data(
            self.vertices_numpy_array, self.edges_numpy_array, self.faces_numpy_array, name
        )

    def to_pymesh(self) -> "PyMeshCookbook":
        from meshemy.cookbook.pymesh import PyMeshCookbook

        if isinstance(self, PyMeshCookbook):
            return self

        return PyMeshCookbook.from_data(self.vertices_numpy_array, self.faces_numpy_array)


T = TypeVar("T")


class MeshIsObjectMixin(GenericModel, Generic[T]):
    mesh: T

    mesh_from_file_loader: ClassVar[Callable[[str], Any]] = ...

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def from_native(cls, mesh: T) -> "MeshIsObjectCookbook":
        return cls(mesh=mesh)

    @classmethod
    def from_file(cls, file_path: FilePath | str) -> "MeshIsObjectCookbook":
        return cls.from_native(mesh=cls.mesh_from_file_loader(str(file_path)))


MeshIsObjectCookbook = TypeVar("MeshIsObjectCookbook", bound=MeshIsObjectMixin)



from functools import partial

from pydantic.generics import GenericModel

try:
    import open3d as o3d
except ModuleNotFoundError as e:
    from meshemy.utility.exception import open3d_module_not_installed_error

    open3d_module_not_installed_error(e)

import logging
from pathlib import Path

import numpy as np
from pydantic_numpy import NDArray, NDArrayFp64

from meshemy.cookbook.base import BaseCookbook, MeshIsObjectMixin
from meshemy.utility.io import o3d_from_vertices_faces
from meshemy.utility.seal import seal_mesh

logger = logging.getLogger(__file__)


class Open3dCookbook(BaseCookbook, MeshIsObjectMixin[o3d.geometry.TriangleMesh]):
    mesh_from_file_loader = o3d.io.read_triangle_mesh

    @property
    def vertices_numpy_array(self) -> NDArrayFp64 | None:
        return np.asarray(self.mesh.vertices)

    @property
    def edges_numpy_array(self) -> NDArray | None:
        return logger.debug("Open3D does not expose the edges of its mesh")

    @property
    def faces_numpy_array(self) -> NDArray | None:
        return np.asarray(self.triangles)

    def smoothen(self, iterations: int, copy: bool = False) -> None:
        result = self.mesh.filter_smooth_taubin(number_of_iterations=iterations)
        if copy:
            self.__class__(mesh=result)
        self.mesh = result

    def attempt_seal_insurance(self) -> bool:
        # TODO: Fix crash
        # Perform seal only if mesh is leaky, ie not watertight
        if not self.watertight:
            logger.debug(f"Attempting to seal mesh")
            self.mesh = o3d_from_vertices_faces(*seal_mesh(self.vertices.copy(), self.faces.copy()))
            if self.mesh.is_watertight():
                logger.debug(f"The seal was success!")
            else:
                logger.debug(f"Failed to seal!")
                return False
        return True

    def repair(self) -> None:
        self.mesh = (
            self.mesh.remove_duplicated_triangles()
            .remove_duplicated_vertices()
            .remove_degenerate_triangles()
            .remove_unreferenced_vertices()
        )

    @property
    def vertices(self) -> NDArray:
        return np.asarray(self.mesh.vertices)

    @property
    def faces(self) -> NDArray:
        return np.asarray(self.mesh.triangles)

    @property
    def triangles(self) -> NDArray:
        return self.faces

    @property
    def watertight(self) -> bool:
        return self.mesh.is_watertight()

    def save(self, save_path: Path | str) -> None:
        o3d.io.write_triangle_mesh(
            str(save_path),
            self.mesh,
            write_vertex_colors=False,
            write_triangle_uvs=False,
        )

    @classmethod
    def from_data(cls, vertices: NDArrayFp64, faces: NDArray) -> "Open3dCookbook":
        return cls(mesh=o3d_from_vertices_faces(vertices, faces))
